@@ -1,8 +1,8 @@
 /*                      _             _
-**  _ __ ___   ___   __| |    ___ ___| |
-** | '_ ` _ \ / _ \ / _` |   / __/ __| |
-** | | | | | | (_) | (_| |   \__ \__ \ | mod_ssl - Apache Interface to OpenSSL
-** |_| |_| |_|\___/ \__,_|___|___/___/_| http://www.engelschall.com/sw/mod_ssl/
+**  _ __ ___   ___   __| |    ___ ___| |  mod_ssl
+** | '_ ` _ \ / _ \ / _` |   / __/ __| |  Apache Interface to OpenSSL
+** | | | | | | (_) | (_| |   \__ \__ \ |  www.modssl.org
+** |_| |_| |_|\___/ \__,_|___|___/___/_|  ftp.modssl.org
 **                      |_____|
 **  mod_ssl.h
 **  Global header
@@ -27,7 +27,7 @@
  *    software must display the following acknowledgment:
  *    "This product includes software developed by
  *     Ralf S. Engelschall <rse@engelschall.com> for use in the
- *     mod_ssl project (http://www.engelschall.com/sw/mod_ssl/)."
+ *     mod_ssl project (http://www.modssl.org/)."
  *
  * 4. The names "mod_ssl" must not be used to endorse or promote
  *    products derived from this software without prior written
@@ -42,7 +42,7 @@
  *    acknowledgment:
  *    "This product includes software developed by
  *     Ralf S. Engelschall <rse@engelschall.com> for use in the
- *     mod_ssl project (http://www.engelschall.com/sw/mod_ssl/)."
+ *     mod_ssl project (http://www.modssl.org/)."
  *
  * THIS SOFTWARE IS PROVIDED BY RALF S. ENGELSCHALL ``AS IS'' AND ANY
  * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
@@ -78,13 +78,13 @@
 #include <sys/stat.h>
 
 /* OpenSSL headers */
-#include <ssl.h>
-#include <err.h>
-#include <x509.h>
-#include <pem.h>
-#include <crypto.h>
-#include <evp.h>
-#include <rand.h>
+#include <openssl/ssl.h>
+#include <openssl/err.h>
+#include <openssl/x509.h>
+#include <openssl/pem.h>
+#include <openssl/crypto.h>
+#include <openssl/evp.h>
+#include <openssl/rand.h>
 
 /* Apache headers */
 #define CORE_PRIVATE
@@ -201,7 +201,6 @@
 /*
  * Defaults for the configuration
  */
-
 #ifndef SSL_SESSION_CACHE_TIMEOUT
 #define SSL_SESSION_CACHE_TIMEOUT  300
 #endif
@@ -297,11 +296,11 @@
 #endif
 
 /*
- * Check for SSLeay/OpenSSL 0.9.0 and below
- * and enabled Extended API (EAPI)
+ * Check for OpenSSL version and whether
+ * Extended API (EAPI) is enabled
  */
-#if SSL_LIBRARY_VERSION < 0x0900
-#error "SSLeay/OpenSSL versions below 0.9.0 are no longer supported"
+#if SSL_LIBRARY_VERSION < 0x00903100
+#error "mod_ssl requires OpenSSL 0.9.3 or higher"
 #endif
 #ifndef EAPI
 #error "mod_ssl requires Extended API (EAPI)"
@@ -323,6 +322,21 @@ typedef struct {
     array_header *aData;
 } ssl_ds_table;
 
+/*
+ * Define the certificate algorithm types
+ */
+
+typedef int ssl_algo_t;
+
+#define SSL_ALGO_UNKNOWN (0)
+#define SSL_ALGO_RSA     (1<<0)
+#define SSL_ALGO_DSA     (1<<1)
+#define SSL_ALGO_ALL     (SSL_ALGO_RSA|SSL_ALGO_DSA)
+
+#define SSL_AIDX_RSA     (0)
+#define SSL_AIDX_DSA     (1)
+#define SSL_AIDX_MAX     (2)
+
 /*
  * Define the SSL options
  */
@@ -331,13 +345,13 @@ typedef struct {
 #define SSL_OPT_COMPATENVVARS  (1<<1)
 #define SSL_OPT_EXPORTCERTDATA (1<<2)
 #define SSL_OPT_FAKEBASICAUTH  (1<<3)
-#define SSL_OPT_ALL            (SSL_OPT_COMPATENVVAR|SSL_OPT_EXPORTCERTDATA|SSL_OPT_FAKEBASICAUTH)
+#define SSL_OPT_STRICTREQUIRE  (1<<4)
+#define SSL_OPT_ALL            (SSL_OPT_COMPATENVVAR|SSL_OPT_EXPORTCERTDATA|SSL_OPT_FAKEBASICAUTH|SSL_OPT_STRICTREQUIRE)
 typedef int ssl_opt_t;
 
 /*
  * Define the SSL Protocol options
  */
-
 #define SSL_PROTOCOL_NONE  (0)
 #define SSL_PROTOCOL_SSLV2 (1<<0)
 #define SSL_PROTOCOL_SSLV3 (1<<1)
@@ -448,6 +462,8 @@ typedef struct {
     int             nInitCount;
     RSA            *pRSATmpKey512;
     RSA            *pRSATmpKey1024;
+    DH             *pDHTmpParam512;
+    DH             *pDHTmpParam1024;
     int             nSessionCacheMode;
     char           *szSessionCacheDataFile;
     ssl_mutexmode_t nMutexMode;
@@ -472,8 +488,8 @@ typedef struct {
  */
 typedef struct {
     BOOL         bEnabled;
-    char        *szCertificateFile;
-    char        *szKeyFile;
+    char        *szPublicCertFile[SSL_AIDX_MAX];
+    char        *szPrivateKeyFile[SSL_AIDX_MAX];
     char        *szCACertificatePath;
     char        *szCACertificateFile;
     char        *szLogFile;
@@ -482,13 +498,16 @@ typedef struct {
     int          nLogLevel;
     int          nVerifyDepth;
     ssl_verify_t nVerifyClient;
-    X509        *px509Certificate;
-    RSA         *prsaKey;
+    X509        *pPublicCert[SSL_AIDX_MAX];
+    EVP_PKEY    *pPrivateKey[SSL_AIDX_MAX];
     SSL_CTX     *pSSLCtx;
     int          nSessionCacheTimeout;
     int          nPassPhraseDialogType;
     char        *szPassPhraseDialogPath;
     ssl_proto_t  nProtocol;
+    char        *szCARevocationPath;
+    char        *szCARevocationFile;
+    X509_STORE  *pRevocationStore;
 #ifdef SSL_VENDOR
     ap_ctx      *ctx;
 #endif
@@ -541,6 +560,8 @@ const char  *ssl_cmd_SSLCertificateFile(cmd_parms *, char *, char *);
 const char  *ssl_cmd_SSLCertificateKeyFile(cmd_parms *, char *, char *);
 const char  *ssl_cmd_SSLCACertificatePath(cmd_parms *, SSLDirConfigRec *, char *);
 const char  *ssl_cmd_SSLCACertificateFile(cmd_parms *, SSLDirConfigRec *, char *);
+const char  *ssl_cmd_SSLCARevocationPath(cmd_parms *, SSLDirConfigRec *, char *);
+const char  *ssl_cmd_SSLCARevocationFile(cmd_parms *, SSLDirConfigRec *, char *);
 const char  *ssl_cmd_SSLVerifyClient(cmd_parms *, SSLDirConfigRec *, char *level);
 const char  *ssl_cmd_SSLVerifyDepth(cmd_parms *, SSLDirConfigRec *, char *);
 const char  *ssl_cmd_SSLSessionCache(cmd_parms *, char *, char *);
@@ -556,7 +577,8 @@ const char  *ssl_cmd_SSLRequire(cmd_parms *, SSLDirConfigRec *, char *);
 void         ssl_init_Module(server_rec *, pool *);
 void         ssl_init_SSLLibrary(void);
 void         ssl_init_GetCertAndKey(server_rec *, pool *, SSLSrvConfigRec *);
-STACK       *ssl_init_FindCAList(server_rec *, pool *, char *, char *);
+STACK_OF(X509_NAME) 
+            *ssl_init_FindCAList(server_rec *, pool *, char *, char *);
 void         ssl_init_Child(server_rec *, pool *);
 void         ssl_init_ModuleKill(void *data);
 
@@ -569,18 +591,17 @@ void         ssl_hook_TimeoutConnection(int);
 void         ssl_hook_CloseConnection(conn_rec *);
 int          ssl_hook_Translate(request_rec *);
 int          ssl_hook_Auth(request_rec *);
+int          ssl_hook_UserCheck(request_rec *);
 int          ssl_hook_Access(request_rec *);
 int          ssl_hook_Fixup(request_rec *);
 int          ssl_hook_ReadReq(request_rec *);
 int          ssl_hook_Handler(request_rec *);
 
 /*  OpenSSL callbacks */
-#if SSL_LIBRARY_VERSION >= 0x0920
 RSA         *ssl_callback_TmpRSA(SSL *, int, int);
-#else
-RSA         *ssl_callback_TmpRSA(SSL *, int);
-#endif
+DH          *ssl_callback_TmpDH(SSL *, int, int);
 int          ssl_callback_SSLVerify(int, X509_STORE_CTX *);
+int          ssl_callback_SSLVerify_CRL(int, X509_STORE_CTX *, server_rec *);
 int          ssl_callback_NewSessionCacheEntry(SSL *, SSL_SESSION *);
 SSL_SESSION *ssl_callback_GetSessionCacheEntry(SSL *, unsigned char *, int, int *);
 void         ssl_callback_DelSessionCacheEntry(SSL_CTX *, SSL_SESSION *);
@@ -603,6 +624,10 @@ void         ssl_scache_dbm_expire(server_rec *);
 void         ssl_pphrase_Handle(server_rec *, pool *);
 int          ssl_pphrase_Handle_CB(char *, int, int);
 
+/*  Diffie-Hellman Parameter Support  */
+DH           *ssl_dh_GetTmpParam(int);
+DH           *ssl_dh_GetParamFromFile(char *);
+
 /*  Data Structures */
 ssl_ds_array *ssl_ds_array_make(pool *, int);
 BOOL          ssl_ds_array_isempty(ssl_ds_array *);
@@ -647,7 +672,7 @@ char        *ssl_var_lookup(pool *, server_rec *, conn_rec *, request_rec *, cha
 /*  I/O  */
 void         ssl_io_register(void);
 void         ssl_io_unregister(void);
-long         ssl_io_data_cb(BIO *, int, char *, int, long, long);
+long         ssl_io_data_cb(BIO *, int, const char *, int, long, long);
 
 /*  PRNG  */
 int          ssl_rand_seed(server_rec *, pool *, ssl_rsctx_t);
@@ -672,6 +697,8 @@ int          ssl_util_ppopen_child(void *, child_info *);
 void         ssl_util_ppclose(server_rec *, pool *, FILE *);
 char        *ssl_util_readfilter(server_rec *, pool *, char *);
 BOOL         ssl_util_path_check(ssl_pathcheck_t, char *);
+ssl_algo_t   ssl_util_algotypeof(X509 *, EVP_PKEY *); 
+char        *ssl_util_algotypestr(ssl_algo_t);
 char        *ssl_util_ptxtsub(pool *, const char *, const char *, char *);
 void         ssl_util_thread_setup(void);
 
