@@ -1,8 +1,8 @@
 /*                      _             _
-**  _ __ ___   ___   __| |    ___ ___| |
-** | '_ ` _ \ / _ \ / _` |   / __/ __| |
-** | | | | | | (_) | (_| |   \__ \__ \ | mod_ssl - Apache Interface to OpenSSL
-** |_| |_| |_|\___/ \__,_|___|___/___/_| http://www.engelschall.com/sw/mod_ssl/
+**  _ __ ___   ___   __| |    ___ ___| |  mod_ssl
+** | '_ ` _ \ / _ \ / _` |   / __/ __| |  Apache Interface to OpenSSL
+** | | | | | | (_) | (_| |   \__ \__ \ |  www.modssl.org
+** |_| |_| |_|\___/ \__,_|___|___/___/_|  ftp.modssl.org
 **                      |_____|
 **  ssl_engine_init.c
 **  Initialization of Servers
@@ -27,7 +27,7 @@
  *    software must display the following acknowledgment:
  *    "This product includes software developed by
  *     Ralf S. Engelschall <rse@engelschall.com> for use in the
- *     mod_ssl project (http://www.engelschall.com/sw/mod_ssl/)."
+ *     mod_ssl project (http://www.modssl.org/)."
  *
  * 4. The names "mod_ssl" must not be used to endorse or promote
  *    products derived from this software without prior written
@@ -42,7 +42,7 @@
  *    acknowledgment:
  *    "This product includes software developed by
  *     Ralf S. Engelschall <rse@engelschall.com> for use in the
- *     mod_ssl project (http://www.engelschall.com/sw/mod_ssl/)."
+ *     mod_ssl project (http://www.modssl.org/)."
  *
  * THIS SOFTWARE IS PROVIDED BY RALF S. ENGELSCHALL ``AS IS'' AND ANY
  * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
@@ -161,7 +161,7 @@ void ssl_init_Module(server_rec *s, pool *p)
         ssl_log_open(s2, p);
     }
 
-    /* 
+    /*
      * Identification
      */
     if (mc->nInitCount == 1)
@@ -170,7 +170,7 @@ void ssl_init_Module(server_rec *s, pool *p)
                 ssl_var_lookup(p, NULL, NULL, NULL, "SSL_VERSION_INTERFACE"),
                 ssl_var_lookup(p, NULL, NULL, NULL, "SSL_VERSION_LIBRARY"));
 
-    /* 
+    /*
      * Initialization round information
      */
     if (mc->nInitCount == 1)
@@ -182,7 +182,7 @@ void ssl_init_Module(server_rec *s, pool *p)
                 mc->nInitCount-2, (mc->nInitCount-2) == 1 ? "st" : "nd");
 
 #ifdef SSL_VENDOR
-    ap_hook_use("ap::mod_ssl::vendor::init_module", 
+    ap_hook_use("ap::mod_ssl::vendor::init_module",
                 AP_HOOK_SIG3(void,ptr,ptr), AP_HOOK_ALL, s, p);
 #endif
 
@@ -222,12 +222,12 @@ void ssl_init_Module(server_rec *s, pool *p)
      */
 
 #ifdef SHARED_MODULE
-    ssl_log(s, SSL_LOG_INFO, "Init: %snitializing %s library", 
+    ssl_log(s, SSL_LOG_INFO, "Init: %snitializing %s library",
             mc->nInitCount == 1 ? "I" : "Rei", SSL_LIBRARY_NAME);
     ssl_init_SSLLibrary();
 #else
     if (mc->nInitCount <= 2) {
-        ssl_log(s, SSL_LOG_INFO, "Init: %snitializing %s library", 
+        ssl_log(s, SSL_LOG_INFO, "Init: %snitializing %s library",
                 mc->nInitCount == 1 ? "I" : "Rei", SSL_LIBRARY_NAME);
         ssl_init_SSLLibrary();
     }
@@ -265,7 +265,7 @@ void ssl_init_Module(server_rec *s, pool *p)
      *  pre-generate the temporary RSA keys
      */
     if (mc->pRSATmpKey512 == NULL) {
-        ssl_log(s, SSL_LOG_INFO, "Init: Generating temporary RSA private keys");
+        ssl_log(s, SSL_LOG_INFO, "Init: Generating temporary RSA private keys (512/1024 bits)");
         mc->pRSATmpKey512 = RSA_generate_key(512, RSA_F4, NULL, NULL);
         if (mc->pRSATmpKey512 == NULL) {
             ssl_log(s, SSL_LOG_ERROR, "Init: Failed to generate temporary 512 bit RSA private key");
@@ -278,6 +278,23 @@ void ssl_init_Module(server_rec *s, pool *p)
         }
     }
 
+    /*
+     *  pre-configure the temporary DH parameters
+     */
+    if (mc->pDHTmpParam512 == NULL) {
+        ssl_log(s, SSL_LOG_INFO, "Init: Configuring temporary DH parameters (512/1024 bits)");
+        mc->pDHTmpParam512 = ssl_dh_GetTmpParam(512);
+        if (mc->pDHTmpParam512 == NULL) {
+            ssl_log(s, SSL_LOG_ERROR, "Init: Failed to configure temporary 512 bit DH parameters");
+            ssl_die();
+        }
+        mc->pDHTmpParam1024 = ssl_dh_GetTmpParam(1024);
+        if (mc->pDHTmpParam1024 == NULL) {
+            ssl_log(s, SSL_LOG_ERROR, "Init: Failed to configure temporary 1024 bit DH parameters");
+            ssl_die();
+        }
+    }
+
     /*
      *  initialize servers
      */
@@ -285,7 +302,7 @@ void ssl_init_Module(server_rec *s, pool *p)
     for (; s != NULL; s = s->next) {
         sc = mySrvConfig(s);
 
-        /* 
+        /*
          * Give out warnings when HTTPS is configured for
          * the HTTP port or vice versa
          */
@@ -298,7 +315,7 @@ void ssl_init_Module(server_rec *s, pool *p)
                     "Init: You configured HTTP(%d) on the standard HTTPS(%d) port!",
                     DEFAULT_HTTP_PORT, DEFAULT_HTTPS_PORT);
 
-        /* 
+        /*
          * Either now skip this server when SSL is disabled for
          * it or give out some information about what we're
          * configuring.
@@ -336,7 +353,7 @@ void ssl_init_SSLLibrary(void)
     CRYPTO_malloc_init();
 #endif
     SSL_load_error_strings();
-    OpenSSL_add_ssl_algorithms();
+    SSL_library_init();
     ssl_util_thread_setup();
     return;
 }
@@ -349,11 +366,13 @@ void ssl_init_GetCertAndKey(server_rec *s, pool *p, SSLSrvConfigRec *sc)
     SSLModConfigRec *mc = myModConfig();
     int nVerify;
     char *cpVHostID;
+    EVP_PKEY *pKey;
     SSL_CTX *ctx;
-    STACK *skCAList;
+    STACK_OF(X509_NAME) *skCAList;
     ssl_asn1_t *asn1;
     unsigned char *ucp;
     char *cp;
+    BOOL ok;
 
     /*
      * Create the server host:port string because we need it a lot
@@ -364,7 +383,7 @@ void ssl_init_GetCertAndKey(server_rec *s, pool *p, SSLSrvConfigRec *sc)
      * Now check for important parameters and the
      * possibility that the user forgot to set them.
      */
-    if (sc->szCertificateFile == NULL) {
+    if (sc->szPublicCertFile[0] == NULL) {
         ssl_log(s, SSL_LOG_ERROR,
                 "Init: (%s) No SSL Certificate set [hint: SSLCertificateFile]",
                 cpVHostID);
@@ -374,7 +393,8 @@ void ssl_init_GetCertAndKey(server_rec *s, pool *p, SSLSrvConfigRec *sc)
     /*
      *  Check for problematic re-initializations
      */
-    if (sc->px509Certificate != NULL) {
+    if (sc->pPublicCert[SSL_AIDX_RSA] != NULL ||
+        sc->pPublicCert[SSL_AIDX_DSA] != NULL   ) {
         ssl_log(s, SSL_LOG_ERROR,
                 "Init: (%s) Illegal attempt to re-initialise SSL for server "
                 "(theoretically shouldn't happen!)", cpVHostID);
@@ -390,14 +410,14 @@ void ssl_init_GetCertAndKey(server_rec *s, pool *p, SSLSrvConfigRec *sc)
                 cpVHostID);
         ssl_die();
     }
-    cp = ap_pstrcat(p, (sc->nProtocol & SSL_PROTOCOL_SSLV2 ? "SSLv2, " : ""), 
-                       (sc->nProtocol & SSL_PROTOCOL_SSLV3 ? "SSLv3, " : ""), 
+    cp = ap_pstrcat(p, (sc->nProtocol & SSL_PROTOCOL_SSLV2 ? "SSLv2, " : ""),
+                       (sc->nProtocol & SSL_PROTOCOL_SSLV3 ? "SSLv3, " : ""),
                        (sc->nProtocol & SSL_PROTOCOL_TLSV1 ? "TLSv1, " : ""), NULL);
     cp[strlen(cp)-2] = NUL;
-    ssl_log(s, SSL_LOG_TRACE, 
+    ssl_log(s, SSL_LOG_TRACE,
             "Init: (%s) Creating new SSL context (protocols: %s)", cpVHostID, cp);
     if (sc->nProtocol == SSL_PROTOCOL_SSLV2)
-        ctx = SSL_CTX_new(SSLv2_server_method());  /* only SSLv2 is left */ 
+        ctx = SSL_CTX_new(SSLv2_server_method());  /* only SSLv2 is left */
     else
         ctx = SSL_CTX_new(SSLv23_server_method()); /* be more flexible */
     SSL_CTX_set_options(ctx, SSL_OP_ALL);
@@ -405,8 +425,9 @@ void ssl_init_GetCertAndKey(server_rec *s, pool *p, SSLSrvConfigRec *sc)
         SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv2);
     if (!(sc->nProtocol & SSL_PROTOCOL_SSLV3))
         SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv3);
-    if (!(sc->nProtocol & SSL_PROTOCOL_TLSV1)) 
+    if (!(sc->nProtocol & SSL_PROTOCOL_TLSV1))
         SSL_CTX_set_options(ctx, SSL_OP_NO_TLSv1);
+    SSL_CTX_set_options(ctx, SSL_OP_SINGLE_DH_USE);
     SSL_CTX_set_app_data(ctx, s);
     sc->pSSLCtx = ctx;
 
@@ -424,6 +445,7 @@ void ssl_init_GetCertAndKey(server_rec *s, pool *p, SSLSrvConfigRec *sc)
     SSL_CTX_sess_set_get_cb(ctx,      ssl_callback_GetSessionCacheEntry);
     SSL_CTX_sess_set_remove_cb(ctx,   ssl_callback_DelSessionCacheEntry);
     SSL_CTX_set_tmp_rsa_callback(ctx, ssl_callback_TmpRSA);
+    SSL_CTX_set_tmp_dh_callback(ctx,  ssl_callback_TmpDH);
     SSL_CTX_set_info_callback(ctx,    ssl_callback_LogTracingState);
 
     /*
@@ -447,7 +469,7 @@ void ssl_init_GetCertAndKey(server_rec *s, pool *p, SSLSrvConfigRec *sc)
         ssl_log(s, SSL_LOG_TRACE,
                 "Init: (%s) Configuring client authentication", cpVHostID);
         if (!SSL_CTX_load_verify_locations(sc->pSSLCtx,
-                                           sc->szCACertificateFile, 
+                                           sc->szCACertificateFile,
                                            sc->szCACertificatePath)) {
             ssl_log(s, SSL_LOG_ERROR|SSL_ADD_SSLERR,
                     "Init: (%s) Unable to configure verify locations "
@@ -464,53 +486,100 @@ void ssl_init_GetCertAndKey(server_rec *s, pool *p, SSLSrvConfigRec *sc)
         SSL_CTX_set_client_CA_list(sc->pSSLCtx, skCAList);
     }
 
+    /*
+     * Configure Certificate Revocation List (CRL) Details
+     */
+    if (sc->szCARevocationFile != NULL || sc->szCARevocationPath != NULL) {
+        ssl_log(s, SSL_LOG_TRACE,
+                "Init: (%s) Configuring certificate revocation facility", cpVHostID);
+        if ((sc->pRevocationStore =
+                SSL_X509_STORE_create(sc->szCARevocationFile,
+                                      sc->szCARevocationPath)) == NULL) {
+            ssl_log(s, SSL_LOG_ERROR|SSL_ADD_SSLERR,
+                    "Init: (%s) Unable to configure X.509 CRL storage "
+                    "for certificate revocation", cpVHostID);
+            ssl_die();
+        }
+    }
+
     /*
      * Give a warning when no CAs were configured but client authentication
      * should take place. This cannot work.
      */
     if (sc->nVerifyClient == SSL_CVERIFY_REQUIRE) {
         skCAList = SSL_CTX_get_client_CA_list(sc->pSSLCtx);
-        if (sk_num(skCAList) == 0)
-            ssl_log(s, SSL_LOG_WARN, 
+        if (sk_X509_NAME_num(skCAList) == 0)
+            ssl_log(s, SSL_LOG_WARN,
                     "Init: Ops, you want to request client authentication, "
                     "but no CAs are known for verification!? "
                     "[Hint: SSLCACertificate*]");
     }
 
     /*
-     *  Configure server certificate
+     *  Configure server certificate(s)
      */
-    ssl_log(s, SSL_LOG_TRACE,
-            "Init: (%s) Configuring server certificate", cpVHostID);
-    if ((asn1 = (ssl_asn1_t *)ssl_ds_table_get(mc->tPublicCert, cpVHostID)) == NULL) {
-        ssl_log(s, SSL_LOG_ERROR,
-                "Init: (%s) Ops, can't find server certificate?!", cpVHostID);
-        ssl_die();
+    ok = FALSE;
+    cp = ap_psprintf(p, "%s:RSA", cpVHostID);
+    if ((asn1 = (ssl_asn1_t *)ssl_ds_table_get(mc->tPublicCert, cp)) != NULL) {
+        ssl_log(s, SSL_LOG_TRACE,
+                "Init: (%s) Configuring RSA server certificate", cpVHostID);
+        ucp = asn1->cpData;
+        sc->pPublicCert[SSL_AIDX_RSA] = d2i_X509(NULL, &ucp, asn1->nData);
+        ok = TRUE;
     }
-    ucp = asn1->cpData;
-    if ((sc->px509Certificate = d2i_X509(NULL, &ucp, asn1->nData)) == NULL) {
+    cp = ap_psprintf(p, "%s:DSA", cpVHostID);
+    if ((asn1 = (ssl_asn1_t *)ssl_ds_table_get(mc->tPublicCert, cp)) != NULL) {
+        ssl_log(s, SSL_LOG_TRACE,
+                "Init: (%s) Configuring DSA server certificate", cpVHostID);
+        ucp = asn1->cpData;
+        sc->pPublicCert[SSL_AIDX_DSA] = d2i_X509(NULL, &ucp, asn1->nData);
+        ok = TRUE;
+    }
+    if (!ok) {
         ssl_log(s, SSL_LOG_ERROR,
-                "Init: (%s) Failed to convert server certificate "
-                "from DER to internal format?!", cpVHostID);
+                "Init: (%s) Ops, no RSA or DSA server certificate found?!", cpVHostID);
         ssl_die();
     }
 
     /*
-     *  Configure server private key
+     *  Configure server private key(s)
      */
-    ssl_log(s, SSL_LOG_TRACE,
-            "Init: (%s) Configuring server private key", cpVHostID);
-    if ((asn1 = (ssl_asn1_t *)ssl_ds_table_get(mc->tPrivateKey, cpVHostID)) == NULL) {
+    ok = FALSE;
+    cp = ap_psprintf(p, "%s:RSA", cpVHostID);
+    if ((asn1 = (ssl_asn1_t *)ssl_ds_table_get(mc->tPrivateKey, cp)) != NULL) {
+        ssl_log(s, SSL_LOG_TRACE,
+                "Init: (%s) Configuring RSA server private key", cpVHostID);
+        ucp = asn1->cpData;
+        sc->pPrivateKey[SSL_AIDX_RSA] = 
+            d2i_PrivateKey(EVP_PKEY_RSA, NULL, &ucp, asn1->nData);
+        ok = TRUE;
+    }
+    cp = ap_psprintf(p, "%s:DSA", cpVHostID);
+    if ((asn1 = (ssl_asn1_t *)ssl_ds_table_get(mc->tPrivateKey, cp)) != NULL) {
+        ssl_log(s, SSL_LOG_TRACE,
+                "Init: (%s) Configuring DSA server private key", cpVHostID);
+        ucp = asn1->cpData;
+        sc->pPrivateKey[SSL_AIDX_DSA] = 
+            d2i_PrivateKey(EVP_PKEY_DSA, NULL, &ucp, asn1->nData);
+        ok = TRUE;
+    }
+    if (!ok) {
         ssl_log(s, SSL_LOG_ERROR,
-                "Init: (%s) Ops, can't find server private key?!", cpVHostID);
+                "Init: (%s) Ops, no RSA or DSA server private key found?!", cpVHostID);
         ssl_die();
     }
-    ucp = asn1->cpData;
-    if ((sc->prsaKey = d2i_RSAPrivateKey(NULL, &ucp, asn1->nData)) == NULL) {
-        ssl_log(s, SSL_LOG_ERROR,
-                "Init: (%s) Failed to convert server private key "
-                "from DER to internal format?!", cpVHostID);
-        ssl_die();
+
+    /*
+     * Optionally copy DSA parameters for certificate from private key
+     * (see http://www.psy.uq.edu.au/~ftp/Crypto/ssleay/TODO.html)
+     */
+    if (   sc->pPublicCert[SSL_AIDX_DSA] != NULL
+        && sc->pPrivateKey[SSL_AIDX_DSA] != NULL) {
+        pKey = X509_get_pubkey(sc->pPublicCert[SSL_AIDX_DSA]);
+        if (   pKey != NULL
+            && EVP_PKEY_type(pKey->type) == EVP_PKEY_DSA 
+            && EVP_PKEY_missing_parameters(pKey))
+            EVP_PKEY_copy_parameters(pKey, sc->pPrivateKey[SSL_AIDX_DSA]);
     }
 
     return;
@@ -521,10 +590,10 @@ static int ssl_init_FindCAList_X509NameCmp(X509_NAME **a, X509_NAME **b)
     return(X509_NAME_cmp(*a, *b));
 }
 
-STACK *ssl_init_FindCAList(server_rec *s, pool *pp, char *cpCAfile, char *cpCApath)
+STACK_OF(X509_NAME) *ssl_init_FindCAList(server_rec *s, pool *pp, char *cpCAfile, char *cpCApath)
 {
-    STACK *skCAList;
-    STACK *sk;
+    STACK_OF(X509_NAME) *skCAList;
+    STACK_OF(X509_NAME) *sk;
     DIR *dir;
     struct DIR_TYPE *direntry;
     char *cp;
@@ -542,19 +611,19 @@ STACK *ssl_init_FindCAList(server_rec *s, pool *pp, char *cpCAfile, char *cpCApa
      * Start with a empty stack/list where new
      * entries get added in sorted order.
      */
-    skCAList = sk_new(ssl_init_FindCAList_X509NameCmp);
+    skCAList = sk_X509_NAME_new(ssl_init_FindCAList_X509NameCmp);
 
     /*
      * Process CA certificate bundle file
      */
     if (cpCAfile != NULL) {
         sk = SSL_load_client_CA_file(cpCAfile);
-        for(n = 0; sk != NULL && n < sk_num(sk); n++) {
+        for(n = 0; sk != NULL && n < sk_X509_NAME_num(sk); n++) {
             ssl_log(s, SSL_LOG_TRACE,
                     "CA certificate: %s",
-                    X509_NAME_oneline((X509_NAME *)sk_value(sk, n), NULL, 0));
-            if (sk_find(skCAList, sk_value(sk, n)) < 0)
-                sk_push(skCAList, sk_value(sk, n));
+                    X509_NAME_oneline(sk_X509_NAME_value(sk, n), NULL, 0));
+            if (sk_X509_NAME_find(skCAList, sk_X509_NAME_value(sk, n)) < 0)
+                sk_X509_NAME_push(skCAList, sk_X509_NAME_value(sk, n));
         }
     }
 
@@ -566,12 +635,12 @@ STACK *ssl_init_FindCAList(server_rec *s, pool *pp, char *cpCAfile, char *cpCApa
         while ((direntry = readdir(dir)) != NULL) {
             cp = ap_pstrcat(p, cpCApath, "/", direntry->d_name, NULL);
             sk = SSL_load_client_CA_file(cp);
-            for(n = 0; sk != NULL && n < sk_num(sk); n++) {
+            for(n = 0; sk != NULL && n < sk_X509_NAME_num(sk); n++) {
                 ssl_log(s, SSL_LOG_TRACE,
                         "CA certificate: %s",
-                        X509_NAME_oneline((X509_NAME *)sk_value(sk, n), NULL, 0));
-                if (sk_find(skCAList, sk_value(sk, n)) < 0)
-                    sk_push(skCAList, sk_value(sk, n));
+                        X509_NAME_oneline(sk_X509_NAME_value(sk, n), NULL, 0));
+                if (sk_X509_NAME_find(skCAList, sk_X509_NAME_value(sk, n)) < 0)
+                    sk_X509_NAME_push(skCAList, sk_X509_NAME_value(sk, n));
             }
         }
         ap_pclosedir(p, dir);
@@ -580,7 +649,7 @@ STACK *ssl_init_FindCAList(server_rec *s, pool *pp, char *cpCAfile, char *cpCApa
     /*
      * Cleanup
      */
-    sk_set_cmp_func(skCAList, NULL);
+    sk_X509_NAME_set_cmp_func(skCAList, NULL);
     ap_destroy_pool(p);
 
     return skCAList;
@@ -605,13 +674,21 @@ void ssl_init_ModuleKill(void *data)
     s = (server_rec *)data;
     for (; s != NULL; s = s->next) {
         sc = mySrvConfig(s);
-        if (sc->px509Certificate != NULL) {
-            X509_free(sc->px509Certificate);
-            sc->px509Certificate = NULL;
+        if (sc->pPublicCert[SSL_AIDX_RSA] != NULL) {
+            X509_free(sc->pPublicCert[SSL_AIDX_RSA]);
+            sc->pPublicCert[SSL_AIDX_RSA] = NULL;
         }
-        if (sc->prsaKey != NULL) {
-            RSA_free(sc->prsaKey);
-            sc->prsaKey = NULL;
+        if (sc->pPublicCert[SSL_AIDX_DSA] != NULL) {
+            X509_free(sc->pPublicCert[SSL_AIDX_DSA]);
+            sc->pPublicCert[SSL_AIDX_DSA] = NULL;
+        }
+        if (sc->pPrivateKey[SSL_AIDX_RSA] != NULL) {
+            EVP_PKEY_free(sc->pPrivateKey[SSL_AIDX_RSA]);
+            sc->pPrivateKey[SSL_AIDX_RSA] = NULL;
+        }
+        if (sc->pPrivateKey[SSL_AIDX_DSA] != NULL) {
+            EVP_PKEY_free(sc->pPrivateKey[SSL_AIDX_DSA]);
+            sc->pPrivateKey[SSL_AIDX_DSA] = NULL;
         }
         if (sc->pSSLCtx != NULL) {
             SSL_CTX_free(sc->pSSLCtx);
