@@ -1,8 +1,8 @@
 /*                      _             _
-**  _ __ ___   ___   __| |    ___ ___| |
-** | '_ ` _ \ / _ \ / _` |   / __/ __| |
-** | | | | | | (_) | (_| |   \__ \__ \ | mod_ssl - Apache Interface to OpenSSL
-** |_| |_| |_|\___/ \__,_|___|___/___/_| http://www.engelschall.com/sw/mod_ssl/
+**  _ __ ___   ___   __| |    ___ ___| |  mod_ssl
+** | '_ ` _ \ / _ \ / _` |   / __/ __| |  Apache Interface to OpenSSL
+** | | | | | | (_) | (_| |   \__ \__ \ |  www.modssl.org
+** |_| |_| |_|\___/ \__,_|___|___/___/_|  ftp.modssl.org
 **                      |_____|
 **  ssl_util_ssl.c
 **  Additional Utility Functions for OpenSSL
@@ -27,7 +27,7 @@
  *    software must display the following acknowledgment:
  *    "This product includes software developed by
  *     Ralf S. Engelschall <rse@engelschall.com> for use in the
- *     mod_ssl project (http://www.engelschall.com/sw/mod_ssl/)."
+ *     mod_ssl project (http://www.modssl.org/)."
  *
  * 4. The names "mod_ssl" must not be used to endorse or promote
  *    products derived from this software without prior written
@@ -42,7 +42,7 @@
  *    acknowledgment:
  *    "This product includes software developed by
  *     Ralf S. Engelschall <rse@engelschall.com> for use in the
- *     mod_ssl project (http://www.engelschall.com/sw/mod_ssl/)."
+ *     mod_ssl project (http://www.modssl.org/)."
  *
  * THIS SOFTWARE IS PROVIDED BY RALF S. ENGELSCHALL ``AS IS'' AND ANY
  * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
@@ -73,9 +73,9 @@ int SSL_get_app_data2_idx(void)
    static int app_data2_idx = -1;
 
    if (app_data2_idx < 0) {
-      app_data2_idx = SSL_get_ex_new_index(0, 
+      app_data2_idx = SSL_get_ex_new_index(0,
            "Second Application Data for SSL", NULL, NULL, NULL);
-      app_data2_idx = SSL_get_ex_new_index(0, 
+      app_data2_idx = SSL_get_ex_new_index(0,
            "Second Application Data for SSL", NULL, NULL, NULL);
    }
    return(app_data2_idx);
@@ -136,14 +136,22 @@ X509 *SSL_read_X509(FILE *fp, X509 **x509, int (*cb)())
     return rc;
 }
 
-RSA *SSL_read_RSAPrivateKey(FILE *fp, RSA **rsa, int (*cb)())
+static EVP_PKEY *d2i_PrivateKey_bio(BIO *bio, EVP_PKEY *key)
 {
-    RSA *rc;
+     return ((EVP_PKEY *)ASN1_d2i_bio(
+             (char *(*)())EVP_PKEY_new, 
+             (char *(*)())d2i_PrivateKey, 
+             (bio), (unsigned char **)(key)));
+}
+
+EVP_PKEY *SSL_read_PrivateKey(FILE *fp, EVP_PKEY **key, int (*cb)())
+{
+    EVP_PKEY *rc;
     BIO *bioS;
     BIO *bioF;
 
     /* 1. try PEM (= DER+Base64+headers) */
-    rc = PEM_read_RSAPrivateKey(fp, rsa, cb);
+    rc = PEM_read_PrivateKey(fp, key, cb);
     if (rc == NULL) {
         /* 2. try DER+Base64 */
         fseek(fp, 0L, SEEK_SET);
@@ -153,7 +161,7 @@ RSA *SSL_read_RSAPrivateKey(FILE *fp, RSA **rsa, int (*cb)())
         if ((bioF = BIO_new(BIO_f_base64())) == NULL)
              return NULL;
         bioS = BIO_push(bioF, bioS);
-        rc = d2i_RSAPrivateKey_bio(bioS, NULL);
+        rc = d2i_PrivateKey_bio(bioS, NULL);
         BIO_free(bioF);
         BIO_free(bioS);
         if (rc == NULL) {
@@ -162,14 +170,14 @@ RSA *SSL_read_RSAPrivateKey(FILE *fp, RSA **rsa, int (*cb)())
             if ((bioS = BIO_new(BIO_s_fd())) == NULL)
                 return NULL;
             BIO_set_fd(bioS, fileno(fp), BIO_NOCLOSE);
-            rc = d2i_RSAPrivateKey_bio(bioS, NULL);
+            rc = d2i_PrivateKey_bio(bioS, NULL);
             BIO_free(bioS);
         }
     }
-    if (rc != NULL && rsa != NULL) {
-        if (*rsa != NULL)
-            RSA_free(*rsa);
-        *rsa = rc;
+    if (rc != NULL && key != NULL) {
+        if (*key != NULL)
+            EVP_PKEY_free(*key);
+        *key = rc;
     }
     return rc;
 }
@@ -200,3 +208,87 @@ int SSL_smart_shutdown(SSL *ssl)
     return rc;
 }
 
+/*  _________________________________________________________________
+**
+**  Certificate Revocation List (CRL) Storage
+**  _________________________________________________________________
+*/
+
+X509_STORE *SSL_X509_STORE_create(char *cpFile, char *cpPath)
+{
+    X509_STORE *pStore;
+    X509_LOOKUP *pLookup;
+
+    if (cpFile == NULL && cpPath == NULL)
+        return NULL;
+    if ((pStore = X509_STORE_new()) == NULL)
+        return NULL;
+    if (cpFile != NULL) {
+        if ((pLookup = X509_STORE_add_lookup(pStore, X509_LOOKUP_file())) == NULL) {
+            X509_STORE_free(pStore);
+            return NULL;
+        }
+        X509_LOOKUP_load_file(pLookup, cpFile, X509_FILETYPE_PEM);
+    }
+    if (cpPath != NULL) {
+        if ((pLookup = X509_STORE_add_lookup(pStore, X509_LOOKUP_hash_dir())) == NULL) {
+            X509_STORE_free(pStore);
+            return NULL;
+        }
+        X509_LOOKUP_add_dir(pLookup, cpPath, X509_FILETYPE_PEM);
+    }
+    return pStore;
+}
+
+int SSL_X509_STORE_lookup(X509_STORE *pStore, int nType,
+                          X509_NAME *pName, X509_OBJECT *pObj)
+{
+    X509_STORE_CTX pStoreCtx;
+    int rc;
+
+    X509_STORE_CTX_init(&pStoreCtx, pStore, NULL, NULL);
+    rc = X509_STORE_get_by_subject(&pStoreCtx, nType, pName, pObj);
+    X509_STORE_CTX_cleanup(&pStoreCtx);
+    return rc;
+}
+
+/*  _________________________________________________________________
+**
+**  Cipher Suite Spec String Creation
+**  _________________________________________________________________
+*/
+
+char *SSL_make_ciphersuite(pool *p, SSL *ssl)
+{
+    STACK_OF(SSL_CIPHER) *sk;
+    SSL_CIPHER *c;
+    int i;
+    int l;
+    char *cpCipherSuite;
+    char *cp;
+
+    if (ssl == NULL) 
+        return "";
+    sk = SSL_get_ciphers(ssl);
+    if (sk == NULL) 
+        return "";
+    l = 0;
+    for (i = 0; i < sk_SSL_CIPHER_num(sk); i++) {
+        c = sk_SSL_CIPHER_value(sk, i);
+        l += strlen(c->name+2+1);
+    }
+    cpCipherSuite = (char *)ap_palloc(p, l+1);
+    cp = cpCipherSuite;
+    for (i = 0; i < sk_SSL_CIPHER_num(sk); i++) {
+        c = sk_SSL_CIPHER_value(sk, i);
+        l = strlen(c->name);
+        memcpy(cp, c->name, l);
+        cp += l;
+        *cp++ = '/';
+        *cp++ = (c->valid == 1 ? '1' : '0');
+        *cp++ = ':';
+    }
+    *cp = NUL;
+    return cpCipherSuite;
+}
+
