@@ -10,8 +10,10 @@ mflags="$2"
 openssl="$3"
 support="$4"
 type="$5"
-crt="$6"
-key="$7"
+algo="$6"
+crt="$7"
+key="$8"
+view="$9"
 
 #   we can operate only inside the Apache 1.3 source
 #   tree and only when mod_ssl+OpenSSL is actually configured.
@@ -31,6 +33,7 @@ fi
 sslcrtdir="../conf/ssl.crt"
 sslcsrdir="../conf/ssl.csr"
 sslkeydir="../conf/ssl.key"
+sslprmdir="../conf/ssl.prm"
 
 #   some optional terminal sequences
 case $TERM in
@@ -52,10 +55,46 @@ esac
 echo "${T_MD}SSL Certificate Generation Utility${T_ME} (mkcert.sh)"
 echo "Copyright (c) 1998 Ralf S. Engelschall, All Rights Reserved."
 
+#   on request view certificates only
+if [ ".$view" != . ]; then
+    if [ -f "$sslcrtdir/ca.crt" -a -f "$sslkeydir/ca.key" ]; then
+        echo ""
+        echo "${T_MD}CA X.509 Certificate${T_ME} [ca.crt]"
+        echo "______________________________________________________________________"
+        $openssl x509 -noout -text -in $sslcrtdir/ca.crt
+        echo ""
+        if [ ".`$openssl x509 -noout -text -in $sslcrtdir/ca.crt | grep "Signature Algorithm" | grep -i RSA`" != . ]; then
+            echo "${T_MD}CA RSA Private Key${T_ME} [ca.key]"
+            echo "______________________________________________________________________"
+            $openssl rsa -noout -text -in $sslkeydir/ca.key
+        else
+            echo "${T_MD}CA DSA Private Key${T_ME} [ca.key]"
+            echo "______________________________________________________________________"
+            $openssl dsa -noout -text -in $sslkeydir/ca.key
+        fi
+    fi
+    if [ -f "$sslcrtdir/server.crt" -a -f "$sslkeydir/server.key" ]; then
+        echo ""
+        echo "${T_MD}Server X.509 Certificate${T_ME} [server.crt]"
+        echo "______________________________________________________________________"
+        $openssl x509 -noout -text -in $sslcrtdir/server.crt
+        echo ""
+        if [ ".`$openssl x509 -noout -text -in $sslcrtdir/server.crt | grep "Signature Algorithm" | grep -i RSA`" != . ]; then
+            echo "${T_MD}Server RSA Private Key${T_ME} [server.key]"
+            echo "______________________________________________________________________"
+            $openssl rsa -noout -text -in $sslkeydir/server.key
+        else
+            echo "${T_MD}Server DSA Private Key${T_ME} [server.key]"
+            echo "______________________________________________________________________"
+            $openssl dsa -noout -text -in $sslkeydir/server.key
+        fi
+    fi
+    exit 0
+fi
+
 #   find some random files
 #   (do not use /dev/random here, because this device 
 #   doesn't work as expected on all platforms)
-echo " + finding random files on your platform"
 randfiles=''
 for file in /var/log/messages /var/adm/messages \
             /kernel /vmunix /vmlinuz \
@@ -69,22 +108,13 @@ for file in /var/log/messages /var/adm/messages \
     fi
 done
 
-#   on-demand compile the ca-fix only
-case $type in
-    test|custom)
-       cd $support
-       if [ ! -f ca-fix ]; then
-           echo " + building ca-fix auxiliary tool"
-           $make $mflags ca-fix >/dev/null 2>&1
-           if [ $? -ne 0 ]; then
-               $make $mflags ca-fix
-               echo "**FAILED"
-               exit 1
-           fi
-       fi
-       cd ..
-       cafix="$support/ca-fix"
-       ;;
+#   canonicalize parameters
+case "x$type" in
+    x ) type=test ;;
+esac
+case "x$algo" in
+    xrsa ) algo=RSA ;;
+    xdsa ) algo=DSA ;;
 esac
 
 #   processing
@@ -95,12 +125,20 @@ case $type in
         echo "${T_MD}Generating self-signed Snake Oil certificate [DUMMY]${T_ME}"
         echo "______________________________________________________________________"
         echo ""
-        cp $sslcrtdir/snakeoil.crt $sslcrtdir/server.crt
-        cp $sslkeydir/snakeoil.key $sslkeydir/server.key
+        if [ ".$algo" = . ]; then
+            algo=RSA
+        fi
+        if [ ".$algo" = .RSA ]; then
+            cp $sslcrtdir/snakeoil-rsa.crt $sslcrtdir/server.crt
+            cp $sslkeydir/snakeoil-rsa.key $sslkeydir/server.key
+        else
+            cp $sslcrtdir/snakeoil-dsa.crt $sslcrtdir/server.crt
+            cp $sslkeydir/snakeoil-dsa.key $sslkeydir/server.key
+        fi
         echo "${T_MD}RESULT: Server Certification Files${T_ME}"
         echo ""
         echo "o  ${T_MD}conf/ssl.key/server.key${T_ME}"
-        echo "   The PEM-encoded RSA private key file which you configure"
+        echo "   The PEM-encoded $algo private key file which you configure"
         echo "   with the 'SSLCertificateKeyFile' directive (automatically done"
         echo "   when you install via APACI). ${T_MD}KEEP THIS FILE PRIVATE!${T_ME}"
         echo ""
@@ -117,23 +155,58 @@ case $type in
         echo ""
         echo "${T_MD}Generating test certificate signed by Snake Oil CA [TEST]${T_ME}"
         echo "WARNING: Do not use this for real-life/production systems"
+        if [ ".$algo" = . ]; then
+            echo "______________________________________________________________________"
+            echo ""
+            echo "${T_MD}STEP 0: Decide the signature algorithm used for certificate${T_ME}"
+            echo "The generated X.509 CA certificate can contain either"
+            echo "RSA or DSA based ingredients. Select the one you want to use."
+            def1=R def2=r def=RSA
+            prompt="Signature Algorithm ((R)SA or (D)SA) [$def1]:"
+            while [ 1 ]; do
+                echo dummy | awk '{ printf("%s", prompt); }' "prompt=$prompt"
+                read algo
+                if [ ".$algo" = ".$def1" -o ".$algo" = ".$def2" -o ".$algo" = . ]; then
+                    algo=$def
+                    break
+                elif [ ".$algo" = ".R" -o ".$algo" = ".r" ]; then
+                    algo=RSA
+                    break
+                elif [ ".$algo" = ".D" -o ".$algo" = ".d" ]; then
+                    algo=DSA
+                    break
+                else
+                    echo "mkcert.sh:Warning: Invalid selection" 1>&2
+                fi
+            done
+        fi
         echo "______________________________________________________________________"
         echo ""
-        echo "${T_MD}STEP 1: Generating RSA private key (1024 bit) [server.key]${T_ME}"
+        echo "${T_MD}STEP 1: Generating $algo private key (1024 bit) [server.key]${T_ME}"
         if [ ! -f $HOME/.rnd ]; then
             touch $HOME/.rnd
         fi
-        if [ ".$randfiles" != . ]; then
-            $openssl genrsa -rand $randfiles \
-                            -out $sslkeydir/server.key \
-                            1024
+        if [ $algo = RSA ]; then
+            if [ ".$randfiles" != . ]; then
+                $openssl genrsa -rand $randfiles -out $sslkeydir/server.key 1024
+            else
+                $openssl genrsa -out $sslkeydir/server.key 1024
+            fi
+            if [ $? -ne 0 ]; then
+                echo "mkcert.sh:Error: Failed to generate RSA private key" 1>&2
+                exit 1
+            fi
         else
-            $openssl genrsa -out $sslkeydir/server.key \
-                            1024
-        fi
-        if [ $? -ne 0 ]; then
-            echo "mkcert.sh:Error: Failed to generate RSA private key" 1>&2
-            exit 1
+            echo "Generating DSA private key via SnakeOil CA DSA parameters"
+            if [ ".$randfiles" != . ]; then
+                $openssl gendsa -rand $randfiles -out $sslkeydir/server.key $sslprmdir/snakeoil-ca-dsa.prm
+            else
+                $openssl gendsa -out $sslkeydir/server.key $sslprmdir/snakeoil-ca-dsa.prm
+            fi
+            if [ $? -ne 0 ]; then
+                echo "mkcert.sh:Error: Failed to generate DSA private key" 1>&2
+                exit 1
+            fi
         fi
         echo "______________________________________________________________________"
         echo ""
@@ -176,64 +249,66 @@ EOT
         echo "${T_MD}STEP 3: Generating X.509 certificate signed by Snake Oil CA [server.crt]${T_ME}"
         echo dummy | awk '{ printf("%s", prompt); }' "prompt=Certificate Version (1 or 3) [3]:"
         read certversion
+        extfile=""
         if [ ".$certversion" = .3 -o ".$certversion" = . ]; then
-            certversion=3
-        else
-            certversion=1
+            extfile="-extfile .mkcert.cfg"
+            cat >.mkcert.cfg <<EOT
+extensions = x509v3
+[ x509v3 ]
+subjectAltName   = email:copy
+nsComment        = "mod_ssl generated test server certificate"
+nsCertType       = server
+EOT
         fi
         if [ ! -f .mkcert.serial ]; then
             echo '01' >.mkcert.serial
         fi
-        $openssl x509 -days 365 \
-                      -CAserial .mkcert.serial \
-                      -CA $sslcrtdir/snakeoil-ca.crt \
-                      -CAkey $sslkeydir/snakeoil-ca.key \
-                      -in $sslcsrdir/server.csr -req \
-                      -out $sslcrtdir/server.crt
+        if [ $algo = RSA ]; then
+            $openssl x509 $extfile \
+                          -days 365 \
+                          -CAserial .mkcert.serial \
+                          -CA $sslcrtdir/snakeoil-ca-rsa.crt \
+                          -CAkey $sslkeydir/snakeoil-ca-rsa.key \
+                          -in $sslcsrdir/server.csr -req \
+                          -out $sslcrtdir/server.crt
+        else
+            $openssl x509 $extfile \
+                          -days 365 \
+                          -CAserial .mkcert.serial \
+                          -CA $sslcrtdir/snakeoil-ca-dsa.crt \
+                          -CAkey $sslkeydir/snakeoil-ca-dsa.key \
+                          -in $sslcsrdir/server.csr -req \
+                          -out $sslcrtdir/server.crt
+        fi
         if [ $? -ne 0 ]; then
             echo "mkcert.sh:Error: Failed to generate X.509 certificate" 1>&2
             exit 1
         fi
-        if [ ".$certversion" = .3 ]; then
-            echo "Converting X.509 v1 to v3 certificate"
-            $cafix -nscertype 0x40 \
-                   -nobscrit \
-                   -nosign \
-                   -in    $sslcrtdir/server.crt \
-                   -inkey $sslkeydir/server.key \
-                   -out   $sslcrtdir/server.crt.fixed
-            if [ $? -ne 0 ]; then
-                echo "mkcert.sh:Error: Failed to fixup X.509 certificate contents" 1>&2
-                exit 1
-            fi
-            cp $sslcrtdir/server.crt.fixed $sslcrtdir/server.crt
-            rm -f $sslcrtdir/server.crt.fixed
-            $cafix -inkey $sslkeydir/snakeoil-ca.key \
-                   -in    $sslcrtdir/server.crt \
-                   -out   $sslcrtdir/server.crt.fixed
-            if [ $? -ne 0 ]; then
-                echo "mkcert.sh:Error: Failed to fixup X.509 certificate signature" 1>&2
-                exit 1
-            fi
-            cp $sslcrtdir/server.crt.fixed $sslcrtdir/server.crt
-            rm -f $sslcrtdir/server.crt.fixed
-        fi
+        rm -f .mkcert.cfg
         echo "Verify: matching certificate & key modulus"
-        modcrt=`$openssl x509 -noout -modulus -in $sslcrtdir/server.crt`
-        modkey=`$openssl rsa  -noout -modulus -in $sslkeydir/server.key`
+        modcrt=`$openssl x509 -noout -modulus -in $sslcrtdir/server.crt | sed -e 's;.*Modulus=;;'`
+        if [ $algo = RSA ]; then
+            modkey=`$openssl rsa -noout -modulus -in $sslkeydir/server.key | sed -e 's;.*Modulus=;;'`
+        else
+            modkey=`$openssl dsa -noout -modulus -in $sslkeydir/server.key | sed -e 's;.*Key=;;'`
+        fi
         if [ ".$modcrt" != ".$modkey" ]; then
             echo "mkcert.sh:Error: Failed to verify modulus on resulting X.509 certificate" 1>&2
             exit 1
         fi
         echo "Verify: matching certificate signature"
-        $openssl verify -CAfile $sslcrtdir/snakeoil-ca.crt $sslcrtdir/server.crt
+        if [ $algo = RSA ]; then
+            $openssl verify -CAfile $sslcrtdir/snakeoil-ca-rsa.crt $sslcrtdir/server.crt
+        else
+            $openssl verify -CAfile $sslcrtdir/snakeoil-ca-dsa.crt $sslcrtdir/server.crt
+        fi
         if [ $? -ne 0 ]; then
             echo "mkcert.sh:Error: Failed to verify signature on resulting X.509 certificate" 1>&2
             exit 1
         fi
         echo "______________________________________________________________________"
         echo ""
-        echo "${T_MD}STEP 4: Enrypting RSA private key with a pass phrase for security [server.key]${T_ME}"
+        echo "${T_MD}STEP 4: Enrypting $algo private key with a pass phrase for security [server.key]${T_ME}"
         echo "The contents of the server.key file (the generated private key) has to be"
         echo "kept secret. So we strongly recommend you to encrypt the server.key file"
         echo "with a Triple-DES cipher and a Pass Phrase."
@@ -250,18 +325,24 @@ EOT
             fi
         done
         if [ ".$rc" = .y ]; then
-            $openssl rsa -des3 \
-                         -in  $sslkeydir/server.key \
-                         -out $sslkeydir/server.key.crypt
+            if [ $algo = RSA ]; then
+                $openssl rsa -des3 \
+                             -in  $sslkeydir/server.key \
+                             -out $sslkeydir/server.key.crypt
+            else
+                $openssl dsa -des3 \
+                             -in  $sslkeydir/server.key \
+                             -out $sslkeydir/server.key.crypt
+            fi
             if [ $? -ne 0 ]; then
-                echo "mkcert.sh:Error: Failed to encrypt RSA private key" 1>&2
+                echo "mkcert.sh:Error: Failed to encrypt $algo private key" 1>&2
                 exit 1
             fi
             cp $sslkeydir/server.key.crypt $sslkeydir/server.key
             rm -f $sslkeydir/server.key.crypt
-            echo "Fine, you're using an encrypted private key."
+            echo "Fine, you're using an encrypted $algo private key."
         else
-            echo "Warning, you're using an unencrypted private key."
+            echo "Warning, you're using an unencrypted $algo private key."
             echo "Please notice this fact and do this on your own risk."
         fi
         echo "______________________________________________________________________"
@@ -269,7 +350,7 @@ EOT
         echo "${T_MD}RESULT: Server Certification Files${T_ME}"
         echo ""
         echo "o  ${T_MD}conf/ssl.key/server.key${T_ME}"
-        echo "   The PEM-encoded RSA private key file which you configure"
+        echo "   The PEM-encoded $algo private key file which you configure"
         echo "   with the 'SSLCertificateKeyFile' directive (automatically done"
         echo "   when you install via APACI). ${T_MD}KEEP THIS FILE PRIVATE!${T_ME}"
         echo ""
@@ -292,23 +373,61 @@ EOT
     custom)
         echo ""
         echo "${T_MD}Generating custom certificate signed by own CA [CUSTOM]${T_ME}"
+        if [ ".$algo" = . ]; then
+            echo "______________________________________________________________________"
+            echo ""
+            echo "${T_MD}STEP 0: Decide the signature algorithm used for certificates${T_ME}"
+            echo "The generated X.509 certificates can contain either"
+            echo "RSA or DSA based ingredients. Select the one you want to use."
+            def1=R def2=r def=RSA
+            prompt="Signature Algorithm ((R)SA or (D)SA) [$def1]:"
+            while [ 1 ]; do
+                echo dummy | awk '{ printf("%s", prompt); }' "prompt=$prompt"
+                read algo
+                if [ ".$algo" = ".$def1" -o ".$algo" = ".$def2" -o ".$algo" = . ]; then
+                    algo=$def
+                    break
+                elif [ ".$algo" = ".R" -o ".$algo" = ".r" ]; then
+                    algo=RSA
+                    break
+                elif [ ".$algo" = ".D" -o ".$algo" = ".d" ]; then
+                    algo=DSA
+                    break
+                else
+                    echo "mkcert.sh:Warning: Invalid selection" 1>&2
+                fi
+            done
+        fi
         echo "______________________________________________________________________"
         echo ""
-        echo "${T_MD}STEP 1: Generating RSA private key for CA (1024 bit) [ca.key]${T_ME}"
+        echo "${T_MD}STEP 1: Generating $algo private key for CA (1024 bit) [ca.key]${T_ME}"
         if [ ! -f $HOME/.rnd ]; then
             touch $HOME/.rnd
         fi
-        if [ ".$randfiles" != . ]; then
-            $openssl genrsa -rand $randfiles \
-                            -out $sslkeydir/ca.key \
-                            1024
+        if [ $algo = RSA ]; then
+            if [ ".$randfiles" != . ]; then
+                $openssl genrsa -rand $randfiles -out $sslkeydir/ca.key 1024
+            else
+                $openssl genrsa -out $sslkeydir/ca.key 1024
+            fi
+            if [ $? -ne 0 ]; then
+                echo "mkcert.sh:Error: Failed to generate RSA private key" 1>&2
+                exit 1
+            fi
         else
-            $openssl genrsa -out $sslkeydir/ca.key \
-                            1024
-        fi
-        if [ $? -ne 0 ]; then
-            echo "mkcert.sh:Error: Failed to generate RSA private key" 1>&2
-            exit 1
+            if [ ".$randfiles" != . ]; then
+                $openssl dsaparam -rand $randfiles -out $sslprmdir/ca.prm 1024
+                echo "Generating DSA private key:"
+                $openssl gendsa   -rand $randfiles -out $sslkeydir/ca.key $sslprmdir/ca.prm
+            else
+                $openssl dsaparam -out $sslprmdir/ca.prm 1024
+                echo "Generating DSA private key:"
+                $openssl gendsa   -out $sslkeydir/ca.key $sslprmdir/ca.prm
+            fi
+            if [ $? -ne 0 ]; then
+                echo "mkcert.sh:Error: Failed to generate DSA private key" 1>&2
+                exit 1
+            fi
         fi
         echo "______________________________________________________________________"
         echo ""
@@ -351,12 +470,20 @@ EOT
         echo "${T_MD}STEP 3: Generating X.509 certificate for CA signed by itself [ca.crt]${T_ME}"
         echo dummy | awk '{ printf("%s", prompt); }' "prompt=Certificate Version (1 or 3) [3]:"
         read certversion
+        extfile=""
         if [ ".$certversion" = .3 -o ".$certversion" = . ]; then
-            certversion=3
-        else
-            certversion=1
+            extfile="-extfile .mkcert.cfg"
+            cat >.mkcert.cfg <<EOT
+extensions = x509v3
+[ x509v3 ]
+subjectAltName   = email:copy
+basicConstraints = CA:true,pathlen:0
+nsComment        = "mod_ssl generated custom CA certificate"
+nsCertType       = sslCA
+EOT
         fi
-        $openssl x509 -days 365 \
+        $openssl x509 $extfile \
+                      -days 365 \
                       -signkey $sslkeydir/ca.key \
                       -in      $sslcsrdir/ca.csr -req \
                       -out     $sslcrtdir/ca.crt
@@ -364,24 +491,14 @@ EOT
             echo "mkcert.sh:Error: Failed to generate self-signed CA certificate" 1>&2
             exit 1
         fi
-        if [ ".$certversion" = .3 ]; then
-            $cafix -caset \
-                   -nscertype 0x07 \
-                   -pathlen 0 \
-                   -nobscrit \
-                   -in    $sslcrtdir/ca.crt \
-                   -inkey $sslkeydir/ca.key \
-                   -out   $sslcrtdir/ca.crt.fixed
-            if [ $? -ne 0 ]; then
-                echo "mkcert.sh:Error: Failed to fixup X.509 CA certificate" 1>&2
-                exit 1
-            fi
-            cp $sslcrtdir/ca.crt.fixed $sslcrtdir/ca.crt
-            rm -f $sslcrtdir/ca.crt.fixed
-        fi
+        rm -f .mkcert.cfg
         echo "Verify: matching certificate & key modulus"
-        modcrt=`$openssl x509 -noout -modulus -in $sslcrtdir/ca.crt`
-        modkey=`$openssl rsa  -noout -modulus -in $sslkeydir/ca.key`
+        modcrt=`$openssl x509 -noout -modulus -in $sslcrtdir/ca.crt | sed -e 's;.*Modulus=;;'`
+        if [ $algo = RSA ]; then
+            modkey=`$openssl rsa -noout -modulus -in $sslkeydir/ca.key | sed -e 's;.*Modulus=;;'`
+        else
+            modkey=`$openssl dsa -noout -modulus -in $sslkeydir/ca.key | sed -e 's;.*Key=;;'`
+        fi
         if [ ".$modcrt" != ".$modkey" ]; then
             echo "mkcert.sh:Error: Failed to verify modulus on resulting X.509 certificate" 1>&2
             exit 1
@@ -394,21 +511,30 @@ EOT
         fi
         echo "______________________________________________________________________"
         echo ""
-        echo "${T_MD}STEP 4: Generating RSA private key for SERVER (1024 bit) [server.key]${T_ME}"
+        echo "${T_MD}STEP 4: Generating $algo private key for SERVER (1024 bit) [server.key]${T_ME}"
         if [ ! -f $HOME/.rnd ]; then
             touch $HOME/.rnd
         fi
-        if [ ".$randfiles" != . ]; then
-            $openssl genrsa -rand $randfiles \
-                            -out $sslkeydir/server.key \
-                            1024
+        if [ $algo = RSA ]; then
+            if [ ".$randfiles" != . ]; then
+                $openssl genrsa -rand $randfiles -out $sslkeydir/server.key 1024
+            else
+                $openssl genrsa -out $sslkeydir/server.key 1024
+            fi
+            if [ $? -ne 0 ]; then
+                echo "mkcert.sh:Error: Failed to generate RSA private key" 1>&2
+                exit 1
+            fi
         else
-            $openssl genrsa -out $sslkeydir/server.key \
-                            1024
-        fi
-        if [ $? -ne 0 ]; then
-            echo "mkcert.sh:Error: Failed to generate RSA private key" 1>&2
-            exit 1
+            if [ ".$randfiles" != . ]; then
+                $openssl gendsa -rand $randfiles -out $sslkeydir/server.key $sslprmdir/ca.prm
+            else
+                $openssl gendsa -out $sslkeydir/server.key $sslprmdir/ca.prm
+            fi
+            if [ $? -ne 0 ]; then
+                echo "mkcert.sh:Error: Failed to generate DSA private key" 1>&2
+                exit 1
+            fi
         fi
         echo "______________________________________________________________________"
         echo ""
@@ -451,15 +577,22 @@ EOT
         echo "${T_MD}STEP 6: Generating X.509 certificate signed by own CA [server.crt]${T_ME}"
         echo dummy | awk '{ printf("%s", prompt); }' "prompt=Certificate Version (1 or 3) [3]:"
         read certversion
+        extfile=""
         if [ ".$certversion" = .3 -o ".$certversion" = . ]; then
-            certversion=3
-        else
-            certversion=1
+            extfile="-extfile .mkcert.cfg"
+            cat >.mkcert.cfg <<EOT
+extensions = x509v3
+[ x509v3 ]
+subjectAltName   = email:copy
+nsComment        = "mod_ssl generated custom server certificate"
+nsCertType       = server
+EOT
         fi
         if [ ! -f .mkcert.serial ]; then
             echo '01' >.mkcert.serial
         fi
-        $openssl x509 -days 365 \
+        $openssl x509 $extfile \
+                      -days 365 \
                       -CAserial .mkcert.serial \
                       -CA    $sslcrtdir/ca.crt \
                       -CAkey $sslkeydir/ca.key \
@@ -469,32 +602,14 @@ EOT
             echo "mkcert.sh:Error: Failed to generate X.509 certificate" 1>&2
             exit 1
         fi
-        if [ ".$certversion" = .3 ]; then
-            $cafix -nscertype 0x40 \
-                   -nobscrit \
-                   -nosign \
-                   -in    $sslcrtdir/server.crt \
-                   -inkey $sslkeydir/server.key \
-                   -out   $sslcrtdir/server.crt.fixed
-            if [ $? -ne 0 ]; then
-                echo "mkcert.sh:Error: Failed to fixup X.509 certificate contents" 1>&2
-                exit 1
-            fi
-            cp $sslcrtdir/server.crt.fixed $sslcrtdir/server.crt
-            rm -f $sslcrtdir/server.crt.fixed
-            $cafix -inkey $sslkeydir/ca.key \
-                   -in    $sslcrtdir/server.crt \
-                   -out   $sslcrtdir/server.crt.fixed
-            if [ $? -ne 0 ]; then
-                echo "mkcert.sh:Error: Failed to fixup X.509 certificate signature" 1>&2
-                exit 1
-            fi
-            cp $sslcrtdir/server.crt.fixed $sslcrtdir/server.crt
-            rm -f $sslcrtdir/server.crt.fixed
-        fi
+        rm -f .mkcert.cfg
         echo "Verify: matching certificate & key modulus"
-        modcrt=`$openssl x509 -noout -modulus -in $sslcrtdir/server.crt`
-        modkey=`$openssl rsa  -noout -modulus -in $sslkeydir/server.key`
+        modcrt=`$openssl x509 -noout -modulus -in $sslcrtdir/server.crt | sed -e 's;.*Modulus=;;'`
+        if [ $algo = RSA ]; then
+            modkey=`$openssl rsa -noout -modulus -in $sslkeydir/server.key | sed -e 's;.*Modulus=;;'`
+        else
+            modkey=`$openssl dsa -noout -modulus -in $sslkeydir/server.key | sed -e 's;.*Key=;;'`
+        fi
         if [ ".$modcrt" != ".$modkey" ]; then
             echo "mkcert.sh:Error: Failed to verify modulus on resulting X.509 certificate" 1>&2
             exit 1
@@ -507,7 +622,7 @@ EOT
         fi
         echo "______________________________________________________________________"
         echo ""
-        echo "${T_MD}STEP 7: Enrypting RSA private key of CA with a pass phrase for security [ca.key]${T_ME}"
+        echo "${T_MD}STEP 7: Enrypting $algo private key of CA with a pass phrase for security [ca.key]${T_ME}"
         echo "The contents of the ca.key file (the generated private key) has to be"
         echo "kept secret. So we strongly recommend you to encrypt the server.key file"
         echo "with a Triple-DES cipher and a Pass Phrase."
@@ -524,11 +639,17 @@ EOT
             fi
         done
         if [ ".$rc" = .y ]; then
-            $openssl rsa -des3 \
-                         -in  $sslkeydir/ca.key \
-                         -out $sslkeydir/ca.key.crypt
+            if [ $algo = RSA ]; then
+                $openssl rsa -des3 \
+                             -in  $sslkeydir/ca.key \
+                             -out $sslkeydir/ca.key.crypt
+            else
+                $openssl dsa -des3 \
+                             -in  $sslkeydir/ca.key \
+                             -out $sslkeydir/ca.key.crypt
+            fi
             if [ $? -ne 0 ]; then
-                echo "mkcert.sh:Error: Failed to encrypt RSA private key" 1>&2
+                echo "mkcert.sh:Error: Failed to encrypt $algo private key" 1>&2
                 exit 1
             fi
             cp $sslkeydir/ca.key.crypt $sslkeydir/ca.key
@@ -540,7 +661,7 @@ EOT
         fi
         echo "______________________________________________________________________"
         echo ""
-        echo "${T_MD}STEP 8: Enrypting RSA private key of SERVER with a pass phrase for security [server.key]${T_ME}"
+        echo "${T_MD}STEP 8: Enrypting $algo private key of SERVER with a pass phrase for security [server.key]${T_ME}"
         echo "The contents of the server.key file (the generated private key) has to be"
         echo "kept secret. So we strongly recommend you to encrypt the server.key file"
         echo "with a Triple-DES cipher and a Pass Phrase."
@@ -557,18 +678,24 @@ EOT
             fi
         done
         if [ ".$rc" = .y ]; then
-            $openssl rsa -des3 \
-                         -in  $sslkeydir/server.key \
-                         -out $sslkeydir/server.key.crypt
+            if [ $algo = RSA ]; then
+                $openssl rsa -des3 \
+                             -in  $sslkeydir/server.key \
+                             -out $sslkeydir/server.key.crypt
+            else
+                $openssl dsa -des3 \
+                             -in  $sslkeydir/server.key \
+                             -out $sslkeydir/server.key.crypt
+            fi
             if [ $? -ne 0 ]; then
-                echo "mkcert.sh:Error: Failed to encrypt RSA private key" 1>&2
+                echo "mkcert.sh:Error: Failed to encrypt $algo private key" 1>&2
                 exit 1
             fi
             cp $sslkeydir/server.key.crypt $sslkeydir/server.key
             rm -f $sslkeydir/server.key.crypt
-            echo "Fine, you're using an encrypted private key."
+            echo "Fine, you're using an encrypted $algo private key."
         else
-            echo "Warning, you're using an unencrypted private key."
+            echo "Warning, you're using an unencrypted $algo private key."
             echo "Please notice this fact and do this on your own risk."
         fi
         echo "______________________________________________________________________"
@@ -576,7 +703,7 @@ EOT
         echo "${T_MD}RESULT: CA and Server Certification Files${T_ME}"
         echo ""
         echo "o  ${T_MD}conf/ssl.key/ca.key${T_ME}"
-        echo "   The PEM-encoded RSA private key file of the CA which you can"
+        echo "   The PEM-encoded $algo private key file of the CA which you can"
         echo "   use to sign other servers or clients. ${T_MD}KEEP THIS FILE PRIVATE!${T_ME}"
         echo ""
         echo "o  ${T_MD}conf/ssl.crt/ca.crt${T_ME}"
@@ -586,7 +713,7 @@ EOT
         echo "   'SSLCACertificateFile' directive."
         echo ""
         echo "o  ${T_MD}conf/ssl.key/server.key${T_ME}"
-        echo "   The PEM-encoded RSA private key file of the server which you configure"
+        echo "   The PEM-encoded $algo private key file of the server which you configure"
         echo "   with the 'SSLCertificateKeyFile' directive (automatically done"
         echo "   when you install via APACI). ${T_MD}KEEP THIS FILE PRIVATE!${T_ME}"
         echo ""
@@ -630,7 +757,7 @@ EOT
             key=$crt
             sed -e '/-----BEGIN CERTIFICATE/,/-----END CERTIFICATE/p' -e '/.*/d' \
                 <$crt >$sslcrtdir/server.crt
-            sed -e '/-----BEGIN RSA PRIVATE KEY/,/-----END RSA PRIVATE KEY/p' -e '/.*/d' \
+            sed -e '/-----BEGIN ... PRIVATE KEY/,/-----END ... PRIVATE KEY/p' -e '/.*/d' \
                 <$key >$sslkeydir/server.key
         fi
         $openssl x509 -noout -in $sslcrtdir/server.crt
@@ -638,10 +765,15 @@ EOT
             echo "mkcert.sh:Error: Failed to check certificate contents: $crt" 1>&2
             exit 1
         fi
+        if [ ".`grep 'PRIVATE KEY' $sslkeydir/server.key | grep RSA`" != . ]; then
+            algo=RSA
+        else
+            algo=DSA
+        fi
         echo "${T_MD}RESULT: Server Certification Files${T_ME}"
         echo ""
         echo "o  ${T_MD}conf/ssl.key/server.key${T_ME}"
-        echo "   The PEM-encoded RSA private key file which you configure"
+        echo "   The PEM-encoded $algo private key file which you configure"
         echo "   with the 'SSLCertificateKeyFile' directive (automatically done"
         echo "   when you install via APACI). ${T_MD}KEEP THIS FILE PRIVATE!${T_ME}"
         echo ""
